import json, sys, os
import importlib
import importlib.util
from datetime import datetime

migration_file_template = """apply_sql = \"""
%s
\"""

rollback_sql = "%s"

migrations = {
    "apply": apply_sql,
    "rollback": rollback_sql
}
"""

empty_migration_template = migration_file_template % ('WRITE your DDL/DML query here', 'WRITE your ROLLBACK query here.')


class MigrationManager:
    def __init__(self, path):
        self.migration_path = path
        sys.path.append(self.migration_path)
        pass


    def module_from_file(self, module_name, file_path):
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module


    def import_migration_files(self):
        for main_dir, subdirs, files in os.walk(self.migration_path):
            print('\n\nprocessing directory: ', main_dir)
            print('subdirectory: ', subdirs)
            if "__pycache__" in main_dir:
                continue
            for f in files:
                print("processing file:: ", f)
                modulename = f.strip('.py')
                module = self.module_from_file(os.path.basename(main_dir) + '_' + modulename , main_dir + os.path.sep + f)

                print(module.migrations)


    def import_single_migration(self, file_name:str):
        """ Imports your migration python file as module for processing.
            The file_name path must be in tableName/fileName.py format
        """
        full_path = self.migration_path + os.path.sep + os.path.sep + file_name
        if os.path.exists(full_path):
            modulename = file_name.strip('.py')
            module = self.module_from_file( modulename.replace(os.path.sep, '_') , full_path)
            # print(module.migrations)
            return module.migrations
        else:
            raise Exception("✅ {} does not exists.\n Please make sure the name of the migration file is correct and it must be in <tableName>/<fileName>.py format."
                    .format(self.migration_path + os.path.sep + os.path.sep + file_name))


    def create_migration_file(self, table_name, file_name):
        new_file_name = self.get_new_file_name(file_name)
        self.create_dir_write_file(table_name, new_file_name, empty_migration_template)

        print("[*] migration file {} has been generated".format(new_file_name) )


    def create_dir_write_file(self, table_name, file_name, content):
        dir_path = self.migration_path + os.path.sep + table_name
        if not os.path.exists(dir_path):
            if not os.path.isdir(dir_path):
                try:
                    os.mkdir(dir_path)
                except Exception as e:
                    raise e

        mg_file = open(dir_path + os.path.sep + file_name, "w+")
        mg_file.write(content)
        mg_file.close()


    def get_new_file_name(self, file_name):
        now = datetime.now()
        time_part = now.strftime("%Y_%m_%d__%H_%M_%S")
        name_temp = "{}_{}.py"
        return name_temp.format(time_part, file_name)


    def get_all_migration_files(self):
        list1 = []
        for main_dir, subdirs, files in os.walk(self.migration_path):
            # print('\n\nprocessing directory: ', main_dir)
            if "__pycache__" in main_dir:
                continue
            for f in files:
                list1.append(main_dir + os.path.sep + f)

        return list1


    def create_migration_file_with_sql(self, table_name:str, create_sql:str):
        file_name = self.get_new_file_name('create_table_' + table_name )

        drop_sql = "DROP TABLE IF EXISTS {} ;".format(table_name)
        content = migration_file_template % (create_sql, drop_sql)
        self.create_dir_write_file(table_name, file_name, content)

        print("file {}{}{} has been created.".format(table_name, os.path.sep, file_name))
        return table_name + os.path.sep + file_name


    def get_pending_migrations(self, db_results):
        """
            Accepts a resultset of migration table to compare with all migration files on file system and resultset.
        """
        processed_files = []
        for file in db_results:
            if file[3] == "complete":
                processed_files.append(file[1])

        #print(processed_files)
        all_files = self.get_all_migration_files()
        base_path_removed = [f.replace(self.migration_path + os.path.sep, '') for f in all_files]
        #print("base files names:: ", base_path_removed)

        return sorted(list(set(base_path_removed) - set(processed_files)) )



# from .configManager import ConfigManager
# from .dbManager import DBManager
# from .migrationManager import MigrationManager
# from .agent import *
# # __all__ = ["ConfigManager", "DBManager", MigrationManager]




import sys, click, json, os, uuid
from .configManager import ConfigManager
from .dbManager import DBManager
from .migrationManager import MigrationManager
from tabulate import tabulate


json_dict = {}
initial_dirs = ["migration"]

def custom_exit(CODE, message="", ex=""):

    print("\n❌ Oooo... snap     \_(-_-)_/  \n"  )
    if ex!="":
        print(ex)
    if message!="":
        print( message, "\n")

    exit(CODE)


def get_cwd():
    print("working directory:: ", os.getcwd())
    return os.getcwd()


def init_setup(dbhost, dbname, dbusername, dbpassword, projectpath):
    """
        This function will check/create the config.json in project root.
        then it'll check/create the revision table in database
    """
    cwd = get_cwd()
    json_dict = { "host": dbhost, "database": dbname, "user": dbusername, "password": dbpassword }

    try:
        CMobj = ConfigManager()
        CMobj.init(json_dict, projectpath)
    except FileExistsError as e:
        custom_exit(0, "", e)
    except Exception as e:
        custom_exit(1, "Unable to initialize rokso.", e)

    config = CMobj.get_config(cwd)

    db = DBManager(config.get("database"))
    db.create_version_table()


def db_status():
    """
    Checks all the migrations processed so far from database
    then checks all pending migrations.
    1. @TODO:: if referring filenames from database entries
        devise a way for OS's directory seperator to access files without a problem.
    """
    cwd = get_cwd()
    try:
        db = DBManager(ConfigManager().get_config(cwd).get("database"))
        cols , data = db.get_database_state()
    except FileNotFoundError as ex:
        custom_exit(1, "It seems the project setup is not complete.\nPlease run `rokso init` first.", ex)

    # get all successful migrations
    completed_migs = list(filter(lambda el: el[3] == "complete", data))

    # get any previous failed migrations
    failed_migs = list(filter(lambda el: el[3] == "error", data))

    print("Last few successful migrations: ")
    print(tabulate(completed_migs[-10:], headers=cols))

    if len(failed_migs) > 0:
        print("\n[❗] However we have detected few failed migrations in the past. \n Please fix them first.\n")
        print(tabulate(failed_migs, headers=cols))
        custom_exit(0)

    mg = MigrationManager(cwd + os.path.sep + 'migration')
    pending_migrations = mg.get_pending_migrations(data)

    if len(pending_migrations) > 0:
        toshow = []
        for pending in pending_migrations:
            toshow.append((pending, 'NA', 'pending'))

        print("\nPending migrations for application: ")
        print(tabulate(toshow, headers=('filename', 'version', 'status')))
        print("\n")
    else:
        print("\nNo new migration to process.\n")


def create_db_migration(tablename, filename):
    mg = MigrationManager(get_cwd() + os.path.sep + 'migration')

    mg.create_migration_file(tablename, filename)


def apply_migration(migration_file_name):
    """
        Checks if any previous migration is in
        @TODO:: check lockings for ALTER statements
        @TODO:: check for all/none or dependencies multi-table entries
    """
    version_no = str(uuid.uuid4())[:8]
    cwd = get_cwd()
    try:
        db = DBManager(ConfigManager().get_config(cwd).get("database"))
        col, data = db.get_database_state()
    except FileNotFoundError as ex:
        custom_exit(1, "It seems the project setup is not complete.\nPlease run `rokso init` first.", ex)

    # get any previous failed migrations
    failed_migs = list(filter(lambda el: el[3] == "error", data))

    failed_files = [f[1] for f in failed_migs]

    mg = MigrationManager(cwd + os.path.sep + 'migration')
    if migration_file_name:
        # if migration file is not in among the previously failed migrations then do not proceed.
        if len(failed_migs) > 0 and migration_file_name != failed_files[0]:
            print("""\n[❗] We have detected some failed migrations which still need to be fixed.
The given migration file name is not same or belongs to the list of below failed migration.
Please fix below files and follow the following order to apply migration. """)
            print(tabulate(failed_migs, headers=col))
            custom_exit(1)

        # process single migration
        sql = mg.import_single_migration(migration_file_name)

        try:
            print("🌀Applying migration file: ", migration_file_name)
            db.apply_migration(sql.get('apply'), migration_file_name, version_no)
        except Exception as ex:
            pass
        finally:
            print("✅ Your database is at revision# {}".format(version_no) )
            print("\n")

    else:
        # checking for failed migration. If present then attempt to migrate them first and do not proceed with new migrations.
        if len(failed_migs) > 0:
            print("""\n[❗] We have detected some failed migrations. Attempting to run following first.\n Once these are successful run `rokso migrate` again to apply new migrations.""")
            print(tabulate(failed_migs, headers=col))
            pending_migrations = failed_files
        else:
            pending_migrations = mg.get_pending_migrations(data)

        if len(pending_migrations) > 0:
            for p_mig in pending_migrations:

                sql = mg.import_single_migration(p_mig)
                try:
                    print("🌀Applying migration file: ", p_mig)
                    db.apply_migration(sql.get('apply'), p_mig, version_no)
                except Exception as ex:
                    print("✅ Your database is at revision# {}".format(version_no) )
                    custom_exit(1, "Your migration '{}' has failed. Please fix it and retry.".format(p_mig), ex)

            print("✅ Your database is at revision# {}".format(version_no) )
            print("\n")
        else:
            print("Nothing to migrate ....\n")


def rollback_db_migration(version):
    # check the revision number in database if nothing is found then report error
    # get all the files for the given revision number and all new revision after that.
    # render details on screen about all the eligible files for rollback
    # on confirmation process rollback one by one.
    cwd = get_cwd()
    try:
        db = DBManager(ConfigManager().get_config(cwd).get("database"))
        mg = MigrationManager(cwd + os.path.sep + 'migration')
    except FileNotFoundError as ex:
        custom_exit(1, "It seems the project setup is not complete.\nPlease run `rokso init` first.", ex)

    if version:
        cols, result = db.get_migrations_more_than_revision(version)
    else:
        cols, result = db.get_latest_db_revision()
        # getting all the files on the latest revision number
        cols, result = db.get_migrations_at_revision(result[0][2] )

    if len(result) < 1:
        custom_exit(1, "No Files to rollback. Probably {} is the latest version.".format(version))

    print("Following files will be rolledback: ")
    print(tabulate(result, headers=cols))
    confirm = input("\nPlease confirm to proceed(y/yes):")
    if confirm.lower() in ['y', 'yes']:
        try:
            for roll in result:
                sql = mg.import_single_migration(roll[1])
                print("\n🔄 Rolling back file:: ", roll[1])
                db.rollback_migration(sql.get('rollback'), roll[0])
        except Exception as ex:
            custom_exit(1, "An error occurred while performing rollback.", ex)

        print("✅ Rollback complete.\n")

    else:
        print("No operation performed.\n")


def reverse_engineer_db():
    """
        1. finds all the tables in database
        2. extract the table definition for each table from DB
        3. create the migration files under "migrations" dir located in project root.
        4. make an entry in version table for that migration
        5. @TODO:: get an optional argument of list of table for which the data should also be dumped.
        6. @TODO:: create logic for stored procedures, functions and triggers.
    """
    cwd = get_cwd()
    mg = MigrationManager(cwd + os.path.sep + 'migration')

    try:
        db = DBManager(ConfigManager().get_config(cwd).get("database"))
        cols , data = db.get_database_state()
    except FileNotFoundError as ex:
        custom_exit(1, "It seems the project setup is not complete.\nPlease run `rokso init` first.", ex)

    if len(data) > 0:
        custom_exit(1, "It seems you already have some rokso migrations in your database. Reverse engineering is possible just after the project setup.")
    else:
        print("Starting reverse engineering")
        version_no = str(uuid.uuid4())[:8]
        headers, data = db.select_query("show tables;")
        print("I found {} tables in the database... ".format(len(data)))
        print(tabulate(data, headers=headers))
        print("\n")

        for table in data:
            print('creating migration for table: ', table[0])
            if table[0] != db.revision_table:
                header, definition = db.get_table_definition(table[0])
                print(definition[0][0])
                # print(definition[0][1])
                migration_file_name = mg.create_migration_file_with_sql(table[0], definition[0][1])
                # now insert into the migration_version table
                db.insert_new_migration(migration_file_name, version_no, "complete" )

        print("✅ Reverse engineering of database complete. \n your database is at revision# {}\n".format(version_no))

def last_success():

    db = DBManager(ConfigManager().get_config(get_cwd()).get("database"))
    try:
        cols , data = db.get_latest_db_revision()
        if len(data) > 0:
            print(data[0][2])
        else:
            print("No last revion detected")
            exit(0)
    except Exception as e:
        #print(e.__class__.__name__)
        if int(str(e).split('(')[0]) == 1146:
            custom_exit(0, "Table does not exist, kindly initate rokso", e)
        else:
            custom_exit(1, "something went wrong", e)




import mysql.connector
from mysql.connector import Error
from .connectionManager import ConnectionManager
from time import time

default_version_table_name = "rokso_db_version"

class DBManager:
    def __init__(self, config):
        self.revision_table = config.get("version_table_name", default_version_table_name)
        self.config = {k: config[k] for k in ('host', 'database', 'user', 'password')}


    def execute_query(self, sql):
        connection = ConnectionManager(self.config).get_connection()
        try:
            cursor = connection.cursor()
            print("\nExecuting>> ", sql )
            tic = time()
            cursor.execute(sql)
            connection.commit()
            toc = time()
            print("query completed successfully..")
            print(">> Time taken: {}secs ".format(round(toc - tic, 4)) )

        except Error as e:
            print("There was an error executing sql:: " + sql, "\n ❌", e)
            raise e
        finally:
            if (connection.is_connected()):
                cursor.close()
                connection.close()


    def select_query(self, sql):
        connection = ConnectionManager(self.config).get_connection()
        try:
            cursor = connection.cursor()
            print("\nExecuting>> ", sql )
            tic = time()
            cursor.execute(sql)
            toc = time()
            print(">> Time taken: {}secs ".format(round(toc - tic, 4)) )
            return cursor.column_names, cursor.fetchall()

        except Error as e:
            print("There was an error executing sql:: " + sql, "\n ❌", e)
        finally:
            if (connection.is_connected()):
                cursor.close()
                connection.close()


    def create_version_table(self):
        """ Creates database version table in the given database """

        sql = """
            CREATE TABLE IF NOT EXISTS {} (
                id INT auto_increment NOT NULL,
                filename varchar(255) NOT NULL,
                version varchar(100) NOT NULL,
                status VARCHAR(20) DEFAULT 'pending' NOT NULL,
                scheduledAt timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
                executedAt timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,

                CONSTRAINT id_PK PRIMARY KEY (id),
                CONSTRAINT filename_UNQ UNIQUE KEY (filename)
            ) ENGINE=InnoDB;
        """
        self.execute_query(sql.format(self.revision_table))
        pass


    def get_database_state(self):
        return self.select_query("SELECT * FROM {}".format(self.revision_table))


    def apply_migration(self, sql, filename, version ):

        try:
            self.execute_query(sql)
            self.insert_new_migration(filename, version, 'complete')
        except Error as e:
            self.insert_new_migration(filename, version, 'error')
            raise e


    def insert_new_migration(self, filename, version, status='pending' ):
        sql = """
                INSERT INTO {}
                (filename, version, status, scheduledAt, executedAt)
                VALUES('{}', '{}', '{}', CURRENT_TIMESTAMP, CURRENT_TIMESTAMP)
                ON DUPLICATE KEY UPDATE status = '{}', version = '{}', executedAt=CURRENT_TIMESTAMP;
            """
        self.execute_query(sql.format(self.revision_table, filename, version, status, status, version))


    def rollback_migration(self, sql, id):
        try:
            self.execute_query(sql)
            self.remove_migration(id)
        except Error as e:
            raise e


    def remove_migration(self, id):
        sql = "DELETE FROM {} WHERE id = {} ;"
        return self.execute_query(sql.format(self.revision_table, id))


    def get_table_definition(self, tablename):
        return self.select_query("SHOW CREATE TABLE {}".format(tablename))


    def get_latest_db_revision(self):
        """ returns """
        sql = "SELECT * from {} ORDER BY id DESC LIMIT 1;"
        return self.select_query(sql.format(self.revision_table))


    def get_migrations_at_revision(self, version):
        sql = """ SELECT * FROM {} WHERE version = '{}' ORDER  BY id desc"""
        return self.select_query(sql.format(self.revision_table, version))

    def get_migrations_more_than_revision(self, version):
        sql = """ SELECT * FROM {} WHERE scheduledAt > (SELECT scheduledAt FROM {} WHERE version = '{}' ORDER  BY id desc LIMIT 1) ORDER BY id DESC; """
        return self.select_query(sql.format(self.revision_table,self.revision_table, version))


import click, sys, os, pathlib
# sys.path.append(pathlib.Path(__file__).parent.absolute())


try:
    from .lib import agent
except ImportError:
    from lib import agent


@click.group()
def cli():
    pass

@click.command('init', short_help='🚀 init your migration project. configures db connection parameters')
@click.option('--projectpath', prompt='Enter path to setup project',
    required=True, envvar='MIG_DB_PROJECT_PATH', help="The path where the project will be setup. rokso can create this directory if not exists.")
@click.option('--dbhost', prompt='Enter database hostname ',
    required=True, envvar='MIG_DB_HOST',
    help="Database host where rokso will connect to.")
@click.option('--dbname', prompt='Enter database name ',
    required=True, envvar='MIG_DB_NAME',
    help="Database name where rokso will apply migrations.")
@click.option('--dbusername', prompt='Enter database username ',
    required=True, envvar='MIG_DB_USER', help="Database username for connecting database.")
@click.option('--dbpassword', prompt='Enter database password',
    required=True, hide_input=True, envvar='MIG_DB_PASSWORD',
    help="Database password for connecting database.")
def init(dbhost, dbname, dbusername, dbpassword, projectpath):
    """This commands configures basic environment variables that are needed to cary out database migrations.
    Make sure the given user has ALTER, ALTER ROUTINE, CREATE, CREATE ROUTINE, DELETE, DROP, EXECUTE,
    INDEX, INSERT, SELECT, SHOW DATABASES, UPDATE privileges.
    """
    agent.init_setup(dbhost, dbname, dbusername, dbpassword, projectpath)


@click.command('status', short_help='✅ checks the current state of database and pending migrations')
def status():
    """ checks the current state of database and pending migrations. It's good to run this before running migrate command. """
    # click.echo('checking database status' + __file__)
    agent.db_status()


@click.command('remap', short_help='🔄 Reverse engineer your DB migrations from existing database.')
def remap():
    """ Reverse engineer your DB migrations from existing database.
     Make sure init command is complete and you have a valid config file and migrations directory setup. """
    click.echo('Starting remapping of existing database for versioning')
    agent.reverse_engineer_db()


@click.command('create', short_help='➕ create a database migration.')
@click.option('--tablename', required=True, prompt='Enter table/procedure/function name that you want to create this migration for.',
            help="The table/procedure/function name for which you want to create the migration.")
@click.option('--filename', required=True, prompt='Enter a file name for this migration.',
            help="Name of the migration file.")
def create(tablename, filename):
    """ Creates a migration template file for specified table/entity name. """
    click.echo('creating a migration ...........')
    agent.create_db_migration(tablename, filename)


@click.command('migrate', short_help='⤴️  Apply all outstanding migrations to database.')
@click.option('--migration', help="Specific migration that needs to be carried out.\nThis option must be of format <tableName>/<fileName> and your file must be under the same path inside migration directory")
def migrate(migration):
    """ Apply all outstanding migrations to database.
    By specifing --migration option you can apply just one single migration. """
    # click.echo('Applying following migrations to database....' + migration)
    agent.apply_migration(migration)


@click.command('rollback', short_help='⤵️  Rollback last applied migration')
@click.option('--version',
                help="Rollbacks database state to specified version.\nThese version numbers can be obtained either from database or by running `rokso status`")
def rollback(version):
    """ Rollback last applied out migration
        By specifing --version option you can rollback to a previous DB state. """
    agent.rollback_db_migration(version)

@click.command('last-success', short_help='⤵️  last successful migration version number')

def last_success():
    agent.last_success()

cli.add_command(init)
cli.add_command(status)
cli.add_command(remap)
cli.add_command(create)
cli.add_command(migrate)
cli.add_command(rollback)
cli.add_command(last_success)


def main():
    return cli()


if __name__ == '__main__':
    main()



import mysql.connector
from mysql.connector import Error


class ConnectionManager:
    def __init__(self, config):
        try:
            self.connection = mysql.connector.connect(**config)
            if self.connection.is_connected():
                db_info = self.connection.get_server_info()
                print("Connected to MySQL Server version ", db_info)

        except Error as e:
            print("Error while connecting to MySQL \n", e)
            raise e


    def get_connection(self):
        return self.connection

    def close_connection(self):
        if (self.connection.is_connected()):
            self.connection.close()




import json, sys, os


class ConfigManager():

    initial_dirs = ["migration"]

    def __init__(self):
        self.config = {}


    def create_initial_dirs(self, working_dir:str):
        print("[#] Generating required dir(s) if not exist")

        for dir in self.initial_dirs:
            dirname = working_dir + os.path.sep + dir
            # print('creating:: ', dirname)
            if not os.path.isdir(dirname):
                try:
                    os.mkdir(dirname)
                except Exception as e:
                    raise e


    def init(self, json_dict:dict, cwd:str):
        # create a new config file in project root

        CONFIG_FILE = cwd + os.path.sep + "config.json"

        print("[*] Checking state of config file in CWD")
        if os.path.exists(CONFIG_FILE):
            raise FileExistsError("Config file already exists")
        else:
            json_data = { "database": json_dict }
            json_object = json.dumps(json_data, indent = 4, sort_keys=True) + "\n"

            with open(CONFIG_FILE, "w") as config_file:
                config_file.write(json_object)
                print("[*] Config file has been created")

            self.create_initial_dirs(cwd)


    def get_config(self, cwd:str):
        CONFIG_FILE = cwd + os.path.sep + "config.json"
        # print('reading config file::', CONFIG_FILE)
        if not bool(self.config):
            with open(CONFIG_FILE, 'r') as f:
                self.config = json.loads(f.read())

        return self.config


from setuptools import setup, find_packages, find_namespace_packages

with open('requirements.txt') as f:
    requirements = f.readlines()

long_description = 'Rokso migration package for mysql database migrations'

setup(
        name ='rokso',
        version ='0.1.15',
        author ='Ankur Pandey',
        author_email ='matrixbegins@gmail.com',
        url ='https://github.com/matrixbegins/rokso-migrations',
        description ='Rokso migration package for mysql database migrations.',
        long_description = long_description,
        long_description_content_type ="text/markdown",
        license ='MIT',
        packages=find_namespace_packages(include=['rokso', 'rokso.*', 'rokso.lib.*']),
        entry_points ={
            'console_scripts': [
                'rokso = rokso.rokso:main'
            ]
        },
        python_requires='>=3.3',
        classifiers = [
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Development Status :: 3 - Alpha",
            "Operating System :: POSIX :: Linux",
            "Operating System :: MacOS",
            "Operating System :: MacOS :: MacOS X",
            "Topic :: Database",
            "Programming Language :: Python :: 3.8",
        ],
        keywords ='MySql, database-migration, python-database-migration',
        install_requires = requirements,
        zip_safe = False
)
